from typing import Optional

from adbc_driver_flightsql import dbapi as adbc_dbapi

from scandal._config import get_scandal_host
from scandal.credentials import Credential


def connect(
    org: Optional[str] = None,
    project: Optional[str] = None,
    auth_token: Optional[str] = None,
    credential: Optional[Credential] = None,
    scandal_host: str = get_scandal_host(),
    **kwargs,
):
    """Connect to Scandal.

    By default,
        - For interactive users we log in using an OAuth2 device code flow to the user's default organization.
        - For non-interactive users, we pick up creds from a credentials chain.
    """
    if not auth_token:
        if not credential:
            raise Exception("Expected auth_token or credential")
        auth_token = credential.get_token()

    # For now, we send the token as the Basic Auth password. This is because the response headers
    # aren't respected when using Bearer auth: https://github.com/apache/arrow-adbc/issues/584
    db_kwargs = {"username": "scandal", "password": auth_token}

    if project:
        # Connect directly to a project
        db_kwargs[_custom_header("x-fulcrum-project")] = project
    if org:
        db_kwargs[_custom_header("x-fulcrum-org")] = org

    connection = adbc_dbapi.connect(f"grpc+tcp://{scandal_host}", db_kwargs)

    connection.fulcrum_organization = org
    connection.fulcrum_project = project

    return connection


def _custom_header(name: str):
    # Construct the name of a custom header to pass to the Flight SQL server
    return f"adbc.flight.sql.rpc.call_header.{name}"



import requests

from scandal.credentials.base import Credential
from scandal.credentials.oauth import wellknown

_token_endpoint = wellknown()["token_endpoint"]


class ClientCredential(Credential):
    """A credential that performs an OAuth2 client credentials flow."""

    def __init__(self, client_id: str, client_secret: str):
        self._id = client_id
        self._secret = client_secret

    def get_token(self):
        resp = requests.post(_token_endpoint, auth=(self._id, self._secret))
        if not resp.ok:
            raise Exception(
                f"Client credentials flow failed {resp.status_code}: {resp.text}"
            )
        return resp["access_token"]



import functools
from typing import Optional

import requests

from scandal._config import get_archimedes_url


@functools.cache
def wellknown(issuer: Optional[str] = None):
    if not issuer:
        issuer = get_archimedes_url()
    if not issuer.endswith("/"):
        issuer = issuer + "/"
    return requests.get(f"{issuer}.well-known/openid-configuration").json()


def check_token(token: str) -> bool:
    """Check the validity of a token.

    We could do this locally for JWTs, but for opaque tokens we need to hit the server.
    For now, we will just always hit the userinfo endpoint.
    """
    return requests.get(
        wellknown()["userinfo_endpoint"], headers={"Authorization": f"Bearer {token}"}
    ).ok



import pyarrow
import sqlalchemy
from sqlalchemy.engine import URL
from sqlalchemy.engine.default import DefaultDialect
from sqlalchemy_bigquery import BigQueryDialect

from scandal.credentials import DeviceCredential
from scandal.dbapi.proxy import ScandalDbApiProxy


class BigQueryScandal(BigQueryDialect):
    def create_connect_args(self, url: URL):
        """The Scandal URI format is <dialect>+scandal://<org>/<project>."""
        org = url.host
        project = url.database

        if not url.username:
            raise Exception("Expected Fulcrum client ID as URI username")

        return [], {
            "org": org,
            "project": project,
            "credential": DeviceCredential(url.username),
        }

    @classmethod
    def dbapi(cls):
        return ScandalDbApiProxy(BigQueryDialect.dbapi())

    def has_table(self, connection, table_name, schema=None):
        return super(DefaultDialect, self).has_table(connection, table_name, schema)

    def get_columns(self, connection, table_name, schema=None, **kw):
        catalog = connection.connection.fulcrum_project
        schema: pyarrow.Schema = connection.connection.adbc_get_table_schema(
            table_name,
            catalog_filter=catalog,
            db_schema_filter=schema,
        )

        columns = []
        for column in schema:
            column: pyarrow.Field
            columns.append(
                {
                    "name": column.name,
                    "nullable": column.nullable,
                    "type": _get_sqla_column_type(column),
                }
            )

        return columns

    def get_table_comment(self, connection, table_name, schema=None, **kw):
        return super(DefaultDialect, self).get_table_comment(
            connection, table_name, schema=schema, **kw
        )

    def get_foreign_keys(self, connection, table_name, schema=None, **kw):
        # We have no support for foreign keys
        return []

    def get_pk_constraint(self, connection, table_name, schema=None, **kw):
        # We have no support for primary keys.
        return {"constrained_columns": []}

    def get_indexes(self, connection, table_name, schema=None, **kw):
        # TODO(ngates): support indexes
        return []

    def get_schema_names(self, connection, **kw):
        """Always use the ADBC metadata API to return schema names."""
        # FIXME(ngates): make use of the info_cache kwargs.
        info = (
            connection.connection.adbc_get_objects(depth="db_schemas")
            .read_all()
            .to_pylist()
        )
        schemas = []
        for catalog in info:
            for db_schema in catalog["catalog_db_schemas"]:
                schemas.append(db_schema["db_schema_name"])
        return schemas

    def get_table_names(self, connection, schema=None, **kw):
        """Always use the ADBC metadata API to return schema names."""
        catalog = connection.connection.fulcrum_project
        info = (
            connection.connection.adbc_get_objects(
                depth="tables",
                catalog_filter=catalog,
                db_schema_filter=schema,
                table_types_filter=["TABLE"],
            )
            .read_all()
            .to_pylist()
        )

        tables = []
        for catalog in info:
            for db_schema in catalog["catalog_db_schemas"]:
                if schema and db_schema["db_schema_name"] != schema:
                    continue
                for db_table in db_schema["db_schema_tables"]:
                    if db_table["table_type"] != "TABLE":
                        continue
                    tables.append(db_table["table_name"])

        return tables

    def get_view_names(self, connection, schema=None, **kw):
        catalog = connection.connection.fulcrum_project
        info = (
            connection.connection.adbc_get_objects(
                depth="tables",
                catalog_filter=catalog,
                db_schema_filter=schema,
                table_types_filter=["VIEW"],
            )
            .read_all()
            .to_pylist()
        )

        views = []
        for catalog in info:
            for db_schema in catalog["catalog_db_schemas"]:
                if schema and db_schema["db_schema_name"] != schema:
                    continue
                for db_table in db_schema["db_schema_tables"]:
                    if db_table["table_type"] != "VIEW":
                        continue
                    views.append(db_table["table_name"])

        return views

    def get_view_definition(self, connection, view_name, schema=None, **kw):
        return super(DefaultDialect, self).get_view_definition(
            connection, view_name, schema, **kw
        )


def _get_sqla_column_type(field: pyarrow.Field):
    _type_map = {
        "date32[day]": sqlalchemy.types.Date,
        "float": sqlalchemy.types.Float,
        "double": sqlalchemy.types.Float,  # Double exists in SQLAlchemy 2.0
        "int32": sqlalchemy.types.Integer,
        "int64": sqlalchemy.types.Integer,
        "string": sqlalchemy.types.String,
        # ... and so on
    }

    try:
        coltype = _type_map[str(field.type)]
    except KeyError:
        sqlalchemy.util.warn(
            "Did not recognize type '%s' of column '%s'" % (field.type, field.name)
        )
        coltype = sqlalchemy.types.NullType
    else:
        # if field.type.endswith("NUMERIC"):
        #     coltype = coltype(precision=field.precision, scale=field.scale)
        # elif field.field_type == "STRING" or field.field_type == "BYTES":
        #     coltype = coltype(field.max_length)
        # elif field.field_type == "RECORD" or field.field_type == "STRUCT":
        #     # FIXME(ngates): probably create our own struct type? Like BigQuery does.
        #     coltype = STRUCT(
        #         *(
        #             (subfield.name, _get_sqla_column_type(subfield))
        #             for subfield in field.fields
        #         )
        #     )
        # else:
        #     coltype = coltype()
        coltype = coltype()

    return coltype



import wrapt

from scandal import dbapi


def instance_method(func):
    """Delegates an instance method to a function."""

    def _func(self, *args, **kwargs):
        return func(*args, **kwargs)

    return _func


class ScandalDbApiProxy(wrapt.ObjectProxy):
    connect = instance_method(dbapi.connect)



from scandal.credentials.base import Credential
from scandal.credentials.client import ClientCredential
from scandal.credentials.device import DeviceCredential

__all__ = ["Credential", "ClientCredential", "DeviceCredential"]



import json
import os.path

CLIENT_IDS = {
    "scandal-py": os.environ.get(
        "SCANDAL_CLIENT_ID", "a358618b-92ea-411f-b8f1-99be0f557da0"
    )
}

CREDS_FILE = os.path.expanduser("~/.config/fulcrum/local.creds.json")
_IS_LOCAL_DEV = os.path.exists(CREDS_FILE)


def get_client_id(service: str = "scandal-py"):
    """For local development, we read the client-id out of a file on disk. Otherwise, it's hard-coded."""
    if _IS_LOCAL_DEV:
        with open(CREDS_FILE) as f:
            creds = json.load(f)
            return creds["clientCreds"][service]["clientId"]
    return CLIENT_IDS[service]


def get_scandal_host():
    if _IS_LOCAL_DEV:
        return "localhost:8443"
    return os.environ.get("SCANDAL_HOST", "crowbar.fly.dev:443")


def get_archimedes_url() -> str:
    if _IS_LOCAL_DEV:
        return "http://localhost:3000"
    return os.environ.get("ARCHIMEDES_URL", "https://archimedes-dev.fly.dev")



"""Scandal implementation of the Database API Specification v2.0.

.. _Python Database API Specification v2.0 (DB-API):
   https://www.python.org/dev/peps/pep-0249/
"""

# Use the ADBC DBAPI module, except for some Scandal overrides.


from adbc_driver_flightsql import dbapi as adbc_dbapi

from scandal.dbapi.connection import connect

apilevel = "2.0"

__all__ = adbc_dbapi.__all__ + ["connect" ""]



import time
import webbrowser

import keyring
import requests

from scandal.credentials.base import Credential
from scandal.credentials.oauth import check_token, wellknown

_device_endpoint = wellknown()["device_authorization_endpoint"]
_token_endpoint = wellknown()["token_endpoint"]
# Usually, device flows should use the token_endpoint. But our flow is hosted on a different server for now.
_device_token_endpoint = wellknown().get("device_token_endpoint", _token_endpoint)


class DeviceCredential(Credential):
    """A credential that triggers a Fulcrum OAuth2 Device Authorization flow."""

    def __init__(self, client_id: str):
        self._client_id = client_id
        self._client_name = f"scandal_{client_id}"

    def get_token(self):
        access_token = keyring.get_password(self._client_name, "access_token")
        refresh_token = keyring.get_password(self._client_name, "refresh_token")

        if access_token:
            if check_token(access_token):
                return access_token

        if refresh_token:
            refresh_resp = requests.post(
                _token_endpoint,
                data={
                    "client_id": self._client_id,
                    "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                    "refresh_token": refresh_token,
                },
            )
            if not refresh_resp.ok:
                print("Failed to refresh Scandal token.")
                keyring.delete_password(self._client_name, "access_token")
                keyring.delete_password(self._client_name, "refresh_token")
                return self.get_token()

        resp = requests.post(
            _device_endpoint,
            data={
                "client_id": self._client_id,
                "scope": "offline_access",
            },
        )
        if not resp.ok:
            raise Exception(f"{resp.status_code}: {resp.text}")
        code_resp = resp.json()

        expiry = time.time() + code_resp["expires_in"]

        print("Enter the following code at this URL: " + code_resp["verification_uri"])
        print(code_resp["user_code"])
        webbrowser.open(code_resp["verification_uri_complete"])

        while time.time() < expiry:
            time.sleep(code_resp["interval"])
            token_resp = requests.post(
                _device_token_endpoint,
                data={
                    "client_id": self._client_id,
                    "grant_type": "urn:ietf:params:oauth:grant-type:device_code",
                    "device_code": code_resp["device_code"],
                },
            )

            if token_resp.ok:
                token_resp = token_resp.json()
                access_token = token_resp["access_token"]
                keyring.set_password(self._client_name, "access_token", access_token)

                refresh_token = token_resp.get("refresh_token", "refresh")
                if refresh_token:
                    keyring.set_password(
                        self._client_name, "refresh_token", refresh_token
                    )

                return access_token

            if token_resp.json()["error"] != "authorization_pending":
                token_resp.raise_for_status()

        raise Exception("Timed out waiting for authorization. Please try again")


def _try_delete_password(service: str, username: str):
    try:
        keyring.delete_password(service, username)
    except Exception:
        pass
